# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['libspn_keras',
 'libspn_keras.config',
 'libspn_keras.constraints',
 'libspn_keras.initializers',
 'libspn_keras.layers',
 'libspn_keras.losses',
 'libspn_keras.math',
 'libspn_keras.metrics',
 'libspn_keras.models',
 'libspn_keras.optimizers']

package_data = \
{'': ['*']}

install_requires = \
['colorlover>=0.3.0,<0.4.0',
 'plotly>=4.9.0,<5.0.0',
 'tensorflow-probability>=0.11,<0.12',
 'tensorflow>=2.3,<3.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.7.0,<2.0.0']}

setup_kwargs = {
    'name': 'libspn-keras',
    'version': '0.5.2',
    'description': 'LibSPN-Keras: A fledxible and scalable library for layer-based building and training of Sum-Product Networks',
    'long_description': '[![Python](https://img.shields.io/pypi/pyversions/libspn-keras.svg?style=plastic)](https://badge.fury.io/py/libspn-keras)\n[![tests](https://github.com/pronobis/libspn-keras/workflows/Tests/badge.svg)](https://github.com/pronobis/libspn-keras/actions?workflow=tests)\n[![PyPI](https://img.shields.io/pypi/v/libspn-keras.svg)](https://pypi.org/project/libspn-keras/)\n[![Read the Docs](https://readthedocs.org/projects/libspn-keras/badge/)](https://libspn-keras.readthedocs.io/)\n\n# LibSPN Keras\nLibSPN Keras is a library for constructing and training Sum-Product Networks. By leveraging the\nKeras framework with a TensorFlow backend, it offers both ease-of-use and scalability. Whereas the\npreviously available `libspn` focused on scalability, `libspn-keras` offers scalability **and**\na straightforward Keras-compatible interface.\n\n![](logo.png "LibSPN Keras logo")\n\n## Contents\n* [API Reference](reference.md)\n\n## Documentation\nThe documentation of the library is hosted on [ReadTheDocs](https://libspn-keras.readthedocs.io/en/latest/index.html).\n\n## What are SPNs?\n\nSum-Product Networks (SPNs) are a probabilistic deep architecture with solid theoretical\nfoundations, which demonstrated state-of-the-art performance in several domains. Yet, surprisingly,\nthere are no mature, general-purpose SPN implementations that would serve as a platform for the\ncommunity of machine learning researchers centered around SPNs. LibSPN Keras is a new\ngeneral-purpose Python library, which aims to become such a platform. The library is designed to\nmake it straightforward and effortless to apply various SPN architectures to large-scale datasets\nand problems. The library achieves scalability and efficiency, thanks to a tight coupling with\nTensorFlow and Keras, two frameworks already in use by a large community of researchers and\ndevelopers in multiple domains.\n\n## Dependencies\nCurrently, LibSPN Keras is tested with `tensorflow>=2.0` and `tensorflow-probability>=0.8.0`.\n\n## Installation\n\n```\npip install libspn-keras\n```\n\n## Note on stability of the repo\nCurrently, the repo is in an alpha state. Hence, one can expect some sporadic breaking changes.\n\n## Feature Overview\n- Gradient based training for generative and discriminative problems\n- Hard EM training for generative problems\n- Hard EM training with unweighted weights for generative problems\n- Soft EM training for generative problems\n- [Deep Generalized Convolutional Sum-Product Networks](https://arxiv.org/abs/1902.06155)\n- SPNs with arbitrary decompositions\n- Fully compatible with Keras and TensorFlow 2.0\n- Input dropout\n- Sum child dropout\n- Image completion\n- Model saving\n- Discrete inputs through an `IndicatorLeaf` node\n- Continuous inputs through `NormalLeaf`, `CauchyLeaf` or `LaplaceLeaf`. Each of these distributions support both\nunivariate as well as *multivariate* inputs.\n\n## Examples / Tutorials\n1. **Benchmark**: `libspn-keras` and Einsum Networks. [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/pronobis/libspn-keras/blob/master/examples/notebooks/Benchmark%20With%20Einsum%20Networks.ipynb)\n2. **Image Classification**: A Deep Generalized Convolutional Sum-Product Network (DGC-SPN). [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/pronobis/libspn-keras/blob/master/examples/notebooks/DGC-SPN%20Image%20Classification.ipynb)\n3. **Image Completion**: A Deep Generalized Convolutional Sum-Product Network (DGC-SPN). [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/pronobis/libspn-keras/blob/master/examples/notebooks/DGC-SPN%20Image%20Completion.ipynb)\n4. **Understanding region SPNs** [![Open In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/pronobis/libspn-keras/blob/master/examples/notebooks/Understanding%20Region%20SPNs.ipynb)\n5. More to come, and if you would like to see a tutorial on anything in particular\nplease raise an issue!\n\nCheck out the way we can build complex DGC-SPNs in a layer-wise fashion:\n```python\nimport libspn_keras as spnk\nfrom tensorflow import keras\n\nspnk.set_default_sum_op(spnk.SumOpGradBackprop())\nspnk.set_default_accumulator_initializer(\n    keras.initializers.TruncatedNormal(stddev=0.5, mean=1.0)\n)\n\nsum_product_network = keras.Sequential([\n  spnk.layers.NormalizeStandardScore(input_shape=(28, 28, 1)),\n  spnk.layers.NormalLeaf(\n      num_components=16,\n      location_trainable=True,\n      location_initializer=keras.initializers.TruncatedNormal(\n          stddev=1.0, mean=0.0)\n  ),\n  # Non-overlapping products\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[2, 2],\n      dilations=[1, 1],\n      kernel_size=[2, 2],\n      padding=\'valid\'\n  ),\n  spnk.layers.Local2DSum(num_sums=16),\n  # Non-overlapping products\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[2, 2],\n      dilations=[1, 1],\n      kernel_size=[2, 2],\n      padding=\'valid\'\n  ),\n  spnk.layers.Local2DSum(num_sums=32),\n  # Overlapping products, starting at dilations [1, 1]\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[1, 1],\n      dilations=[1, 1],\n      kernel_size=[2, 2],\n      padding=\'full\'\n  ),\n  spnk.layers.Local2DSum(num_sums=32),\n  # Overlapping products, with dilations [2, 2] and full padding\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[1, 1],\n      dilations=[2, 2],\n      kernel_size=[2, 2],\n      padding=\'full\'\n  ),\n  spnk.layers.Local2DSum(num_sums=64),\n  # Overlapping products, with dilations [2, 2] and full padding\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[1, 1],\n      dilations=[4, 4],\n      kernel_size=[2, 2],\n      padding=\'full\'\n  ),\n  spnk.layers.Local2DSum(num_sums=64),\n  # Overlapping products, with dilations [2, 2] and \'final\' padding to combine\n  # all scopes\n  spnk.layers.Conv2DProduct(\n      depthwise=True,\n      strides=[1, 1],\n      dilations=[8, 8],\n      kernel_size=[2, 2],\n      padding=\'final\'\n  ),\n  spnk.layers.SpatialToRegions(),\n  # Class roots\n  spnk.layers.DenseSum(num_sums=10),\n  spnk.layers.RootSum(return_weighted_child_logits=True)\n])\n\nsum_product_network.summary(line_length=100)\n```\n\nWhich produces:\n```\nModel: "sequential"\n_________________________________________________________________\nLayer (type)                 Output Shape              Param #\n=================================================================\nnormal_leaf (NormalLeaf)     (None, 28, 28, 16)        25088\n_________________________________________________________________\nconv2d_product (Conv2DProduc (None, 14, 14, 16)        4\n_________________________________________________________________\nlocal2d_sum (Local2DSum)     (None, 14, 14, 16)        50176\n_________________________________________________________________\nconv2d_product_1 (Conv2DProd (None, 7, 7, 16)          4\n_________________________________________________________________\nlocal2d_sum_1 (Local2DSum)   (None, 7, 7, 32)          25088\n_________________________________________________________________\nconv2d_product_2 (Conv2DProd (None, 8, 8, 32)          4\n_________________________________________________________________\nlocal2d_sum_2 (Local2DSum)   (None, 8, 8, 32)          65536\n_________________________________________________________________\nconv2d_product_3 (Conv2DProd (None, 10, 10, 32)        4\n_________________________________________________________________\nlocal2d_sum_3 (Local2DSum)   (None, 10, 10, 64)        204800\n_________________________________________________________________\nconv2d_product_4 (Conv2DProd (None, 14, 14, 64)        4\n_________________________________________________________________\nlocal2d_sum_4 (Local2DSum)   (None, 14, 14, 64)        802816\n_________________________________________________________________\nconv2d_product_5 (Conv2DProd (None, 8, 8, 64)          4\n_________________________________________________________________\nspatial_to_regions (SpatialT (None, 1, 1, 4096)        0\n_________________________________________________________________\ndense_sum (DenseSum)         (None, 1, 1, 10)          40960\n_________________________________________________________________\nroot_sum (RootSum)           (None, 10)                10\n=================================================================\nTotal params: 1,214,498\nTrainable params: 1,201,930\nNon-trainable params: 12,568\n_________________________________________________________________\n```\n\n## TODOs\n- Structure learning\n- Advanced regularization e.g. pruning or auxiliary losses on weight accumulators\n',
    'author': 'Jos van de Wolfshaar',
    'author_email': 'jos.vandewolfshaar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pronobis/libspn-keras',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from tensorflow import keras


class OnlineExpectationMaximization(keras.optimizers.SGD):
    """
    Online expectation maximization which requires sum layers to use any of the EM-based SumOpBase instances.

    Internally, this is just an SGD optimizer with unit learning rate.
    """

    def __init__(
        self, learning_rate: float = 0.05, gliding_average: bool = True, **kwargs
    ):
        if gliding_average:
            learning_rate = -learning_rate / (learning_rate - 1.0)
        super().__init__(learning_rate, **kwargs)



import abc
import functools
from typing import Callable, Optional, Tuple, Union

import tensorflow as tf

from libspn_keras.math.logconv import logconv1x1_2d
from libspn_keras.math.logmatmul import logmatmul


def _batch_scope_tranpose(f):  # type: ignore  # noqa: ANN001,ANN202
    @functools.wraps(f)  # type: ignore  # noqa: ANN202
    def impl(self: SumOpBase, x: tf.Tensor, *args, **kwargs) -> tf.Tensor:  # type: ignore
        with tf.name_scope("ScopesAndDecompsFirst"):
            scopes_decomps_first = tf.transpose(x, (1, 2, 0, 3))
        result = f(self, scopes_decomps_first, *args, **kwargs)
        with tf.name_scope("BatchFirst"):
            return tf.transpose(result, (2, 0, 1, 3))

    return impl


class SumOpBase(abc.ABC):
    """
    The base sum op primitive.

    Descendants define weighted sum implementations to override gradients if necessary.
    """

    @abc.abstractmethod
    def weighted_sum(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Implement sum operation on log inputs X and accumulators w.

        Args:
            x: Input Tensor.
            accumulators: Unnormalized accumulators.
            logspace_accumulators: Whether accumulators are in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.
        """

    @abc.abstractmethod
    def weighted_children(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted children (used in RootSum).

        Args:
            x: Input Tensor.
            accumulators: Unnormalized accumulators.
            logspace_accumulators: Whether accumulators are in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.
        """

    @abc.abstractmethod
    def weighted_conv(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted convolution (used in ConvSum).

        Args:
            x: Input Tensor.
            accumulators: Unnormalized accumulators.
            logspace_accumulators: Whether accumulators are in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.
        """

    @abc.abstractmethod
    def default_logspace_accumulators(self) -> bool:
        """Whether default config is to have accumulators in logspace."""

    @staticmethod
    def _to_log_weights(x: tf.Tensor) -> tf.Tensor:
        with tf.name_scope("ToLogWeights"):
            return SumOpBase._log_normalize(tf.math.log(x))

    @staticmethod
    def _log_normalize(x: tf.Tensor) -> tf.Tensor:
        with tf.name_scope("LogNormalize"):
            return tf.nn.log_softmax(x, axis=-2)

    def _to_logspace_override_grad(
        self, accumulators: tf.Tensor, normalize_in_forward_pass: bool
    ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], tf.Tensor]]:
        @tf.custom_gradient
        def _inner(
            accumulators: tf.Tensor,
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], tf.Tensor]]:
            if normalize_in_forward_pass:
                return (
                    self._to_log_weights(accumulators),
                    lambda dy: dy
                    / (tf.abs(tf.reduce_sum(dy, axis=2, keepdims=True)) + 1e-20),
                )
            else:
                return (
                    tf.math.log(accumulators),
                    lambda dy: dy
                    / (tf.abs(tf.reduce_sum(dy, axis=2, keepdims=True)) + 1e-20),
                )

        return _inner(accumulators)

    def _weights_in_logspace(
        self,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        if logspace_accumulators:
            return (
                self._log_normalize(accumulators)
                if normalize_in_forward_pass
                else accumulators
            )
        else:
            return (
                self._to_log_weights(accumulators)
                if normalize_in_forward_pass
                else tf.math.log(accumulators)
            )


class SumOpGradBackprop(SumOpBase):
    """
    Sum op primitive with gradient in backpropagation when computed through TensorFlow's autograd engine.

    Internally, weighted sums are computed with default gradients for all ops being used.

    Args:
            logspace_accumulators: If provided overrides default log-space choice. For a
                ``SumOpGradBackprop`` the default is ``True``
    """

    def __init__(
        self, logspace_accumulators: Optional[bool] = None,
    ):
        self._logspace_accumulators = logspace_accumulators

    @_batch_scope_tranpose
    def weighted_sum(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute a weighted sum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.
        """
        w = self._weights_in_logspace(
            accumulators, logspace_accumulators, normalize_in_forward_pass
        )
        return logmatmul(x, w)

    def weighted_children(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted children, without summing over the final axis.

        This is used for a RootSum to compute :math:`P(X,Y_i)` for any :math:`i`

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.
        """
        w = self._weights_in_logspace(
            accumulators, logspace_accumulators, normalize_in_forward_pass
        )
        return x + tf.linalg.matrix_transpose(w)

    def weighted_conv(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted convolutions.

        This is used for a Conv2DSum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted convolutions.
        """
        w = self._weights_in_logspace(
            accumulators, logspace_accumulators, normalize_in_forward_pass
        )
        return logconv1x1_2d(x, w)

    def default_logspace_accumulators(self) -> bool:
        """
        Whether or not accumulators should be represented in log-space by default.

        Returns:
            True if the default representation is in logspace and False otherwise.
        """
        return (
            self._logspace_accumulators
            if self._logspace_accumulators is not None
            else True
        )


class SumOpEMBackprop(SumOpBase):
    """
    Sum op primitive with EM signals in backpropagation.

    These are dense EM signals as opposed to the other EM based instances of ``SumOpBase``.
    """

    @_batch_scope_tranpose
    def weighted_sum(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute a weighted sum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "EM is only implemented for linear space accumulators"
            )
        w = self._to_logspace_override_grad(accumulators, normalize_in_forward_pass)
        return logmatmul(x, w)

    def weighted_children(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted children, without summing over the final axis.

        This is used for a RootSum to compute :math:`P(X,Y_i)` for any :math:`i`

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "EM is only implemented for linear space accumulators"
            )
        w = self._to_logspace_override_grad(accumulators, normalize_in_forward_pass)
        with tf.name_scope("PairwiseLogMultiply"):
            return x + tf.linalg.matrix_transpose(w)

    def weighted_conv(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted convolutions.

        This is used for a Conv2DSum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted convolutions.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "EM is only implemented for linear space accumulators"
            )
        w = self._to_logspace_override_grad(accumulators, normalize_in_forward_pass)
        return logconv1x1_2d(x, w)

    def default_logspace_accumulators(self) -> bool:
        """
        Whether or not accumulators should be represented in log-space by default.

        Returns:
            True if the default representation is in logspace and False otherwise.
        """
        return False


class SumOpHardEMBackprop(SumOpBase):
    """
    Sum op with hard EM signals in backpropagation when computed through TensorFlow's autograd engine.

    Args:
        sample_prob: Sampling probability in the range of [0, 1]. Sampling logits are taken from
            the normalized log probability of the children of each sum.
    """

    def __init__(self, sample_prob: Optional[Union[float, tf.Tensor]] = None):
        self.sample_prob = sample_prob

    @_batch_scope_tranpose
    def weighted_sum(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute a weighted sum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "Hard EM is only implemented for linear space accumulators"
            )

        @tf.custom_gradient
        def _inner_fn(
            x: tf.Tensor, accumulators: tf.Tensor
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
            with tf.name_scope("HardEMForwardPass"):
                w = (
                    self._to_log_weights(accumulators)
                    if normalize_in_forward_pass
                    else tf.math.log(accumulators)
                )
                # Pairwise product in forward pass
                x = tf.expand_dims(x, axis=3)
                w = tf.expand_dims(tf.linalg.matrix_transpose(w), axis=2)

                # Max per sum for determining winning child + choosing the constant for numerical
                # stability
                weighted_children = x + w
                max_weighted_child = tf.stop_gradient(
                    tf.reduce_max(weighted_children, axis=-1, keepdims=True)
                )

                # Perform log(sum(exp(...))) with the numerical stability trick
                out = tf.math.log(
                    tf.reduce_sum(
                        tf.exp(weighted_children - max_weighted_child), axis=-1
                    )
                ) + tf.squeeze(max_weighted_child, axis=-1)

            def grad(dy: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
                # Determine winning child
                equal_to_max: tf.Tensor = tf.cast(
                    tf.equal(weighted_children, max_weighted_child), tf.float32
                )

                # Holds the index of the winning child per sum
                if self.sample_prob is not None:
                    equal_to_max = (
                        tf.math.log(
                            self.sample_prob * tf.exp(x - max_weighted_child)
                            + tf.convert_to_tensor(1.0 - self.sample_prob)
                            * equal_to_max
                        )
                        + max_weighted_child
                    )
                else:
                    equal_to_max = tf.math.log(equal_to_max)
                num_in = tf.shape(x)[-1]
                equal_to_max_flat_outer = tf.reshape(
                    equal_to_max, tf.concat([[-1], [num_in]], axis=0)
                )

                # Holds the index of the winning child per sum
                winning_child_per_sum = tf.reshape(
                    tf.random.categorical(
                        tf.math.log(equal_to_max_flat_outer), num_samples=1
                    ),
                    tf.shape(out),
                )

                # Pass on the counts to the edges between child and parent
                per_sample_weight_counts = tf.expand_dims(dy, -1) * tf.one_hot(
                    winning_child_per_sum, depth=num_in
                )

                child_counts = tf.reduce_sum(per_sample_weight_counts, axis=3)
                weight_counts = tf.reduce_sum(per_sample_weight_counts, axis=2)

                return child_counts, tf.transpose(weight_counts, (0, 1, 3, 2))

            return out, grad

        return _inner_fn(x, accumulators)

    def weighted_children(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted children, without summing over the final axis.

        This is used for a RootSum to compute :math:`P(X,Y_i)` for any :math:`i`

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "EM is only implemented for linear space accumulators"
            )
        w = self._to_logspace_override_grad(accumulators, normalize_in_forward_pass)
        with tf.name_scope("PairwiseLogMultiply"):
            return x + tf.linalg.matrix_transpose(w)

    @_batch_scope_tranpose
    def weighted_conv(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted convolutions.

        This is used for a Conv2DSum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted convolutions.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "Hard EM is only implemented for linear space accumulators"
            )

        @tf.custom_gradient
        def _inner_fn(
            x: tf.Tensor, accumulators: tf.Tensor
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:
            with tf.name_scope("HardEMForwardPass"):
                w = (
                    self._to_log_weights(accumulators)
                    if normalize_in_forward_pass
                    else tf.math.log(accumulators)
                )
                # Pairwise product in forward pass
                x = tf.expand_dims(x, axis=3)
                w = tf.expand_dims(tf.linalg.matrix_transpose(w), axis=2)

                # Max per sum for determining winning child + choosing the constant for numerical
                # stability
                weighted_children = x + w
                max_weighted_child = tf.stop_gradient(
                    tf.reduce_max(weighted_children, axis=-1, keepdims=True)
                )

                # Perform log(sum(exp(...))) with the numerical stability trick
                out = tf.math.log(
                    tf.reduce_sum(
                        tf.exp(weighted_children - max_weighted_child), axis=-1
                    )
                ) + tf.squeeze(max_weighted_child, axis=-1)

            def grad(dy: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
                # Determine winning child
                equal_to_max = tf.cast(
                    tf.equal(weighted_children, max_weighted_child), tf.float32
                )

                # Holds the index of the winning child per sum
                if self.sample_prob is not None:
                    equal_to_max = (
                        tf.math.log(
                            self.sample_prob * tf.exp(x - max_weighted_child)
                            + (1.0 - self.sample_prob) * equal_to_max
                        )
                        + max_weighted_child
                    )
                else:
                    equal_to_max = tf.math.log(equal_to_max)
                num_in = tf.shape(x)[-1]
                equal_to_max_flat_outer = tf.reshape(
                    equal_to_max, tf.concat([[-1], [num_in]], axis=0)
                )

                # Holds the index of the winning child per sum
                winning_child_per_sum = tf.reshape(
                    tf.random.categorical(
                        tf.math.log(equal_to_max_flat_outer), num_samples=1
                    ),
                    tf.shape(out),
                )

                # Pass on the counts to the edges between child and parent
                per_sample_weight_counts = tf.expand_dims(dy, -1) * tf.one_hot(
                    winning_child_per_sum, depth=num_in
                )

                child_counts = tf.reduce_sum(per_sample_weight_counts, axis=3)
                weight_counts = tf.reduce_sum(per_sample_weight_counts, axis=2)

                # Sum over spatial axes
                weight_counts = tf.reduce_sum(weight_counts, axis=[0, 1], keepdims=True)

                return child_counts, tf.linalg.matrix_transpose(weight_counts)

            return out, grad

        return _inner_fn(x, accumulators)

    def default_logspace_accumulators(self) -> bool:
        """
        Whether or not accumulators should be represented in log-space by default.

        Returns:
            True if the default representation is in logspace and False otherwise.
        """
        return False


class SumOpUnweightedHardEMBackprop(SumOpBase):
    """
    Sum op with hard EM signals in backpropagation when computed through TensorFlow's autograd engine.

    Instead of using weighted sum inputs to select the maximum child, it relies on unweighted child
    inputs, which has the advantage of alleviating a self-amplifying chain of hard EM signals in
    deep SPNs.

    Args:
        sample_prob: Sampling probability in the range of [0, 1]. Sampling logits are taken from
            the normalized log probability of the children of each sum.
    """

    def __init__(self, sample_prob: Optional[Union[float, tf.Tensor]] = None):
        self.sample_prob = sample_prob

    @_batch_scope_tranpose
    def weighted_sum(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute a weighted sum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "Hard EM is only implemented for linear space accumulators"
            )

        @tf.custom_gradient
        def _inner_fn(
            x: tf.Tensor, accumulators: tf.Tensor
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:

            # Normalized
            weights = (
                self._to_log_weights(accumulators)
                if normalize_in_forward_pass
                else tf.math.log(accumulators)
            )

            out = logmatmul(x, weights)

            def grad(parent_counts: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
                # Determine winning child
                max_child = tf.reduce_max(x, axis=-1, keepdims=True)
                equal_to_max = tf.cast(tf.equal(max_child, x), tf.float32)

                # Holds the index of the winning child per sum
                if self.sample_prob is not None:
                    equal_to_max = (
                        tf.math.log(
                            self.sample_prob * tf.exp(x - max_child)
                            + (1.0 - self.sample_prob) * equal_to_max
                        )
                        + max_child
                    )
                else:
                    equal_to_max = tf.math.log(equal_to_max)

                num_in = tf.shape(x)[-1]
                equal_to_max_flat_outer = tf.reshape(
                    equal_to_max, tf.concat([[-1], [num_in]], axis=0)
                )
                winning_child_per_scope = tf.reshape(
                    tf.random.categorical(equal_to_max_flat_outer, num_samples=1),
                    tf.shape(x)[:-1],
                )

                sum_parent_counts = tf.reduce_sum(parent_counts, axis=-1, keepdims=True)

                winning_child_per_scope_one_hot = tf.one_hot(
                    winning_child_per_scope, depth=num_in, axis=-1
                )
                child_counts = winning_child_per_scope_one_hot * sum_parent_counts

                weight_counts = tf.matmul(
                    winning_child_per_scope_one_hot, parent_counts, transpose_a=True
                )
                return child_counts, weight_counts

            return out, grad

        return _inner_fn(x, accumulators)

    def weighted_children(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted children, without summing over the final axis.

        This is used for a RootSum to compute :math:`P(X,Y_i)` for any :math:`i`

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted sums.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "EM is only implemented for linear space accumulators"
            )
        w = self._to_logspace_override_grad(accumulators, normalize_in_forward_pass)
        with tf.name_scope("PairwiseLogMultiply"):
            return x + tf.linalg.matrix_transpose(w)

    def weighted_conv(
        self,
        x: tf.Tensor,
        accumulators: tf.Tensor,
        logspace_accumulators: bool,
        normalize_in_forward_pass: bool,
    ) -> tf.Tensor:
        """
        Compute weighted convolutions.

        This is used for a Conv2DSum.

        Args:
            x: Input Tensor
            accumulators: Accumulators, can be seen as unnormalized representations of weights.
            logspace_accumulators: Whether or not accumulators are represented in logspace.
            normalize_in_forward_pass: Whether weights should be normalized during forward inference.

        Returns:
            A Tensor with the weighted convolutions.

        Raises:
            NotImplementedError: When called with ``losgpace_accumulators == True``.
        """
        if logspace_accumulators:
            raise NotImplementedError(
                "Hard EM is only implemented for linear space accumulators"
            )

        @tf.custom_gradient
        def _inner_fn(
            x: tf.Tensor, accumulators: tf.Tensor
        ) -> Tuple[tf.Tensor, Callable[[tf.Tensor], Tuple[tf.Tensor, tf.Tensor]]]:

            # Normalized
            weights = (
                self._to_log_weights(accumulators)
                if normalize_in_forward_pass
                else tf.math.log(accumulators)
            )

            out = logmatmul(x, weights)

            def grad(parent_counts: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
                # Determine winning child
                max_child = tf.reduce_max(x, axis=-1, keepdims=True)
                equal_to_max = tf.cast(tf.equal(max_child, x), tf.float32)

                # Holds the index of the winning child per sum
                if self.sample_prob is not None:
                    equal_to_max = (
                        tf.math.log(
                            self.sample_prob * tf.exp(x - max_child)
                            + (1.0 - self.sample_prob) * equal_to_max
                        )
                        + max_child
                    )
                else:
                    equal_to_max = tf.math.log(equal_to_max)

                num_in = tf.shape(x)[-1]
                equal_to_max_flat_outer = tf.reshape(
                    equal_to_max, tf.concat([[-1], [num_in]], axis=0)
                )
                winning_child_per_scope = tf.reshape(
                    tf.random.categorical(equal_to_max_flat_outer, num_samples=1),
                    tf.shape(x)[:-1],
                )

                sum_parent_counts = tf.reduce_sum(parent_counts, axis=-1, keepdims=True)

                winning_child_per_scope_one_hot = tf.one_hot(
                    winning_child_per_scope, depth=num_in, axis=-1
                )
                child_counts = winning_child_per_scope_one_hot * sum_parent_counts

                weight_counts = tf.matmul(
                    winning_child_per_scope_one_hot, parent_counts, transpose_a=True
                )
                weight_counts = tf.reduce_sum(weight_counts, axis=[0, 1], keepdims=True)
                return child_counts, weight_counts

            return out, grad

        return _inner_fn(x, accumulators)

    def default_logspace_accumulators(self) -> bool:
        """
        Whether or not accumulators should be represented in log-space by default.

        Returns:
            True if the default representation is in logspace and False otherwise.
        """
        return False



from typing import Optional, Tuple

import tensorflow as tf
from tensorflow import keras


class ReduceProduct(keras.layers.Layer):
    """
    Computes products per decomposition and scope by reduction.

    Assumes the incoming tensor is of shape ``[num_batch, num_scopes, num_decomps, num_nodes]``
    and produces an output of ``[num_batch, num_scopes // num_factors, num_decomps, num_nodes]``.

    Args:
        num_factors: Number of factors per product
        **kwargs: kwargs to pass on to the keras.Layer super class.
    """

    def __init__(self, num_factors: int, **kwargs):
        super(ReduceProduct, self).__init__(**kwargs)
        self.num_factors = num_factors

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: When shape could not be determined.
        """
        _, self._num_scopes_in, self._num_decomps, self._num_nodes_in = input_shape
        if self._num_scopes_in is None:
            raise ValueError("Cannot build product with unknown number of input scopes")
        if self._num_scopes_in % self.num_factors != 0:
            raise ValueError("Number of input scopes is not divisible by factor")
        self._num_scopes = self._num_scopes_in // self.num_factors
        self._num_products = self._num_nodes_in
        super(ReduceProduct, self).build(input_shape)

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute product by reduction over scope subset.

        Args:
            x: Region input tensor.
            kwargs: Remaining keyword arguments.

        Returns:
            Products where by reducing over neighboring scopes.
        """
        # Split in list of tensors which will be added up using outer products
        shape = [
            -1,
            self._num_scopes,
            self.num_factors,
            self._num_decomps,
            self._num_nodes_in,
        ]
        return tf.reduce_sum(tf.reshape(x, shape=shape), axis=2)

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.

        Raises:
            ValueError: If shape cannot be determined.
        """
        num_scopes_in, num_decomps, num_batch, num_nodes_in = input_shape
        if num_scopes_in is None:
            raise ValueError("Cannot compute shape with unknown number of input scopes")
        return (
            num_batch,
            num_scopes_in // self.num_factors,
            self._num_decomps,
            num_nodes_in,
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(num_factors=self.num_factors,)
        base_config = super(ReduceProduct, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Optional, Tuple

import tensorflow as tf

from libspn_keras.layers.dense_sum import DenseSum
from libspn_keras.sum_ops import SumOpBase


class RootSum(DenseSum):
    """
    Final sum of an SPN. Expects input to be in log-space and produces log-space output.

    Args:
        return_weighted_child_logits: If True, returns a weighted child log probability, which
            can be used for e.g. (Sparse)CategoricalCrossEntropy losses. If False, computes
            the weighted sum of the input, which effectively is the log probability of the
            distribution defined by the SPN.
        logspace_accumulators: If ``True``, accumulators will be represented in log-space which
            is typically used with ``BackpropMode.GRADIENT``. If ``False``, accumulators will be
            represented in linear space. Weights are computed by normalizing the accumulators
            per sum, so that we always end up with a normalized SPN. If ``None`` (default) it
            will be set to ``True`` for ``BackpropMode.GRADIENT`` and ``False`` otherwise.
        accumulator_initializer: Initializer for accumulator. If None, defaults to
            initializers.Constant(1.0)
        accumulator_regularizer: Regularizer for accumulator.
        linear_accumulator_constraint: Constraint for linear accumulators. Defaults to a
            constraint that ensures a minimum of a small positive constant. If
            logspace_accumulators is set to True, this constraint wil be ignored
        sum_op (SumOpBase): SumOpBase instance which determines how to compute the forward and
            backward pass of the weighted sums
        **kwargs: kwargs to pass on to the keras.Layer super class
    """

    def __init__(
        self,
        return_weighted_child_logits: bool = True,
        logspace_accumulators: Optional[bool] = None,
        accumulator_initializer: Optional[tf.keras.initializers.Initializer] = None,
        trainable: bool = True,
        logspace_accumulator_constraint: Optional[
            tf.keras.constraints.Constraint
        ] = None,
        accumulator_regularizer: Optional[tf.keras.regularizers.Regularizer] = None,
        linear_accumulator_constraint: Optional[tf.keras.constraints.Constraint] = None,
        sum_op: Optional[SumOpBase] = None,
        **kwargs,
    ):
        super(RootSum, self).__init__(
            logspace_accumulators=logspace_accumulators,
            accumulator_initializer=accumulator_initializer,
            trainable=trainable,
            logspace_accumulator_constraint=logspace_accumulator_constraint,
            accumulator_regularizer=accumulator_regularizer,
            linear_accumulator_constraint=linear_accumulator_constraint,
            sum_op=sum_op,
            num_sums=1,
            **kwargs,
        )
        self.return_weighted_child_logits = return_weighted_child_logits

    def call(self, x: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Compute weighted sum at the root.

        If ``return_weighted_child_logits`` is set to ``True``, this will return P(X,Y_i) for all i
        rather than P(X).

        Args:
            x: Input, must have only 1 decomposition and 1 scope.
            kwargs: Remaining keyword arguments.

        Returns:
            Sum or weighted children.
        """
        x_reshaped = tf.reshape(x, (-1, 1, 1, self._num_nodes_in))

        if self.return_weighted_child_logits:
            out = self.sum_op.weighted_children(
                x_reshaped,
                accumulators=self._accumulators,
                logspace_accumulators=self.logspace_accumulators,
                normalize_in_forward_pass=self._forward_normalize,
            )
            num_out = self._accumulators.shape[2]
        else:
            out = super(RootSum, self).call(x_reshaped, **kwargs)
            num_out = 1
        return tf.reshape(out, [-1, num_out])

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """
        Build the internal components for this layer.

        Args:
            input_shape: Shape of the input Tensor.

        Raises:
            ValueError: Raised when the number of input scopes does not equal 1, or
                if number of decompositions or nodes in the input is unknown.
        """
        num_batch, num_scopes, num_decomps, num_nodes = input_shape
        if num_scopes != 1:
            raise ValueError(
                f"Expected to have 1 scope at the input of a RootSum, got {num_scopes}"
            )
        if num_decomps is None:
            raise ValueError("Must have known number of decompositions")
        if num_nodes is None:
            raise ValueError("Must have known number of nodes")
        super(RootSum, self).build((num_batch, num_scopes, 1, num_decomps * num_nodes))

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.


        Raises:
            ValueError: Raised when the number of decompositions
                or nodes in the input is unknown.
        """
        num_batch, _, num_decomps, num_nodes_in = input_shape
        if num_decomps is None:
            raise ValueError("Must have known number of decompositions")
        if num_nodes_in is None:
            raise ValueError("Must have known number of nodes")
        return (
            num_batch,
            num_decomps * num_nodes_in
            if self.return_weighted_child_logits
            else num_batch,
        )

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(return_weighted_child_logits=self.return_weighted_child_logits)
        base_config = super(RootSum, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



from typing import Optional

import tensorflow as tf
from tensorflow import keras


class LogLikelihood(keras.metrics.Mean):
    r"""
    Compute log marginal :math:`1/N \sum \log(p(X))`.

    Assumes that the last layer of the SPN is a ``RootSum``. It ignores the ``y_true`` argument,
    since a target for :math:`Y` is absent in unsupervised learning.
    """

    def __init__(self, name: str = "llh", **kwargs):
        super(LogLikelihood, self).__init__(name=name, **kwargs)

    def update_state(
        self, _: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[float] = None
    ) -> tf.Tensor:
        """
        Accumulates statistics for computing the reduction metric.

        For example, if `values` is [1, 3, 5, 7] and reduction=SUM_OVER_BATCH_SIZE,
        then the value of `result()` is 4. If the `sample_weight` is specified as
        [1, 1, 0, 0] then value of `result()` would be 2.

        Args:
            y_pred: Predictions coming from the model.
            sample_weight: Optional weighting of each example. Defaults to 1.

        Returns:
            Update op.
        """
        values = tf.reduce_logsumexp(y_pred, axis=-1)
        return super(LogLikelihood, self).update_state(
            values, sample_weight=sample_weight
        )



from typing import Optional, Tuple, Union

import tensorflow as tf
from tensorflow import keras


class NormalizeStandardScore(keras.layers.Layer):
    """
    Normalizes samples to a standard score.

    In other words, the output is the input minus its mean and divided by the standard deviation.
    This can be used to achieve the same kind of normalization as used in (Poon and Domingos, 2011).

    Args:
        normalization_epsilon: Small positive constant to prevent division by zero,
            but could also be used a 'smoothing' factor.
        sample_wise: If ``True``, will compute z-scores where statistics are computed sample-wise, otherwise,
            computes z-scores through computing cross-sample statistics. To use cross-sample statistics, call
            ``NormalizeStandardScore.adapt(train_ds)`` where ``train_ds`` is an instance of
            ``tf.data.Dataset``.
        **kwargs: kwargs to pass on to the keras.Layer super class

    References:
        Sum-Product Networks, a New Deep Architecture
        `Poon and Domingos, 2011 <https://arxiv.org/abs/1202.3732>`_
    """

    def __init__(
        self, normalization_epsilon: float = 1e-8, sample_wise: bool = True, **kwargs
    ):
        super(NormalizeStandardScore, self).__init__(**kwargs)
        self.normalization_epsilon = normalization_epsilon
        self.sample_wise = sample_wise
        self.mean = None
        self.stddev = None

    def adapt(self, ds: tf.data.Dataset) -> None:
        """
        Compute cross-sample statistics. Assumes that ``ds`` is a batched dataset.

        Args:
            ds: Instance of ``tf.data.Dataset`` containing train data to compute statistics from.
        """
        shape = tf.shape(next(ds.take(1).as_numpy_iterator())[0])
        sum_x = tf.zeros(shape)
        sum_x2 = tf.zeros(shape)
        count = tf.constant(0.0)

        for x in ds:
            sum_x += tf.reduce_sum(tf.cast(x, tf.float32), axis=0)
            sum_x2 += tf.reduce_sum(tf.square(tf.cast(x, tf.float32)), axis=0)
            count += tf.cast(tf.shape(x)[0], tf.float32)

        self.mean = self.add_weight(
            "mean",
            initializer=tf.keras.initializers.Constant(sum_x / count),
            trainable=False,
            shape=shape,
        )
        self.stddev = self.add_weight(
            "stddev",
            initializer=tf.keras.initializers.Constant(
                tf.sqrt(sum_x2 / count - tf.square(sum_x / count))
            ),
            shape=shape,
            trainable=False,
        )

    def call(
        self, x: tf.Tensor, return_stats: bool = False, **kwargs
    ) -> Union[Tuple[tf.Tensor, ...], tf.Tensor]:
        """
        Normalize raw input by subtracting mean and dividing by the standard deviation per sample.

        Args:
            x: Raw input tensor.
            return_stats: Whether to return mean and standard edviation
            kwargs: Remaining keyword arguments.

        Returns:
            A normalized tensor or a tuple of the normalized tensor, the mean and the standard deviation.

        Raises:
            ValueError: if ``NoramalizeStandardScore.adapt(train_ds)`` wasn't called before calling this
                layer's forward pass.
        """
        if self.sample_wise:
            normalization_axes_indices = tf.range(1, tf.rank(x))
            mean = tf.reduce_mean(x, axis=normalization_axes_indices, keepdims=True)
            stddev = tf.math.reduce_std(
                x, axis=normalization_axes_indices, keepdims=True
            )
        else:
            if self.mean is None or self.stddev is None:
                raise ValueError(
                    "Cannot compute normalized values before calling adapt()"
                )
            mean = self.mean
            stddev = self.stddev
        normalized_input = (x - mean) / (stddev + self.normalization_epsilon)
        if return_stats:
            return normalized_input, mean, stddev
        return normalized_input

    def compute_output_shape(
        self, input_shape: Tuple[Optional[int], ...]
    ) -> Tuple[Optional[int], ...]:
        """
        Compute output shape of the layer.

        Args:
            input_shape: Input shape of the layer.

        Returns:
            Tuple of ints holding the output shape of the layer.
        """
        return input_shape

    def get_config(self) -> dict:
        """
        Obtain a key-value representation of the layer config.

        Returns:
            A dict holding the configuration of the layer.
        """
        config = dict(normalization_epsilon=self.normalization_epsilon)
        base_config = super(NormalizeStandardScore, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))
